<?php
namespace Util\Events\Task;

use Doctrine\Common\EventSubscriber;
use Doctrine\ORM\Event\LifecycleEventArgs;
use Doctrine\ORM\Event\PostFlushEventArgs;
use Doctrine\ORM\Event\PreUpdateEventArgs;
use Doctrine\ORM\Events;

use Util\PmCache;
use Util\Utilities;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\Appropriation;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\CalendarDates;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Log\TaskLogInterface;
use Zeedhi\DevOps\Model\Projectmanagement\Listeners\Project;
use Zeedhi\DevOps\Model\Projectmanagement\Listeners\Task;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Log\ProjectLogInterface;
use Zeedhi\Framework\DependencyInjection\InstanceManager;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Calendar\TaskCalendarInterface;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Tree\Treeble;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\Task as TaskEntity;
use Zeedhi\DevOps\Model\Projectmanagement\Entities\Project as ProjectEntity;
use Zeedhi\DevOps\Model\Projectmanagement\Entities;

class Subscriber implements EventSubscriber {

    const SQL_TOCHILDREN = "SELECT TREE_KEY
  FROM
  TASK P
  WHERE TREE_KEY like :KEY AND TREE_KEY NOT LIKE :KEY_PARENT
  AND STATUS <> 'D'
    UNION
SELECT TREE_KEY
  FROM
  PROJECT P
  WHERE TREE_KEY like :KEY AND TREE_KEY NOT LIKE :KEY_PARENT
  AND STATUS <> 'D'";

	/** @var \Service\Queue\Task\Producer $taskQueueProducer */
	protected $taskQueueProducer;
	protected $oldParent;
	protected $oldProject;
	protected $logProject;
    protected $needsFlush;
    protected $logIssue;

    /***********************************************************************
     * Constructor.
     ***********************************************************************/
	public function __construct($taskQueueProducer) {
		$this->taskQueueProducer = $taskQueueProducer;
        $this->logProject = [];
        $this->logIssue = [];
	}

    /***********************************************************************
     * Updates an entity, since tasks needs to have its tree code updated.
     **********************************************************************
     * @param Entities\Task | Entities\Project $entity
     * @param $e
     */
	private function updateEntity($entity, $e){
        if($entity->hasChildren()) {
            /** @var Entities\Task[] | Entities\Project[] $children */
            $children = $entity->children();
        }else{
            $children = null;
        }
		$entity->setTreeCode($this->nextId($entity->getTreeParent(), $entity->getTableName(), $e));
        if(is_array($children)){
            foreach($children as $child){
                $child->setTreeCode($this->nextId($child->getParent(), $child->getTableName(), $e));
            }
        }
	}

    /***********************************************************************
     * Check if is the entity an instance of TaskCalendarInterface
     ***********************************************************************/
    private function isEntityTaskCalendarInterface($entity){
        return $entity instanceof TaskCalendarInterface;
    }

    private function isEntityCalendarDate($entity){
        return $entity instanceof CalendarDates;
    }

    /***********************************************************************
     * Check if is the entity an instance of Entities\Task
     ***********************************************************************/
    private function isEntityTask($entity){
        return $entity instanceof TaskEntity;
    }

    /***********************************************************************
     * Check if is the entity an instance of Treeble
     ***********************************************************************/
    private function isEntityTreeble($entity){
        return $entity instanceof Treeble;
    }

    /***********************************************************************
     * Check if is the entity an instance of Appropriation
     ***********************************************************************/
    private function isEntityAppropriation($entity){
        return $entity instanceof Appropriation;
    }

    /***********************************************************************
     * Check if is the entity an instance of Allocation
     ***********************************************************************/
    private function isEntityAllocation($entity){
        return $entity instanceof Allocation;
    }

    /***********************************************************************
     * Check if is the entity an instance of ProjectLogInterface
     ***********************************************************************/
    private function isEntityProjectLogInterface($entity){
        return $entity instanceof ProjectLogInterface;
    }

    /***********************************************************************
     * Check if is the entity an instance of Project
     ***********************************************************************/
    private function isEntityProject($entity){
        return $entity instanceof ProjectEntity;
    }

    /***********************************************************************
     * Doctrine event. Called before a data is created in a database.
     ***********************************************************************/
	public function prePersist(LifecycleEventArgs $e){
        /** @var Entities\Task | Entities\Project $entity */
		$entity = $e->getEntity();

		if ($this->isEntityTreeble($entity)){
            /** @var \Service\Calc\TaskCalculatorImpl $calculator */
			$calculator = InstanceManager::getInstance()->getService('taskCalculator');
            $entity->setTreeCode($this->nextId($entity->getTreeParent(), $entity->getTableName(), $e));
            $calculator->calculate($entity);
        }
		if ($this->isEntityAppropriation($entity)){
			$calculator = InstanceManager::getInstance()->getService('taskCalculator');
			$calculator->calculate($entity->getTask());
		}
	}

    /***********************************************************************
     * Doctrine event. Called after a data is created in a database.
     ***********************************************************************/
	public function postPersist(LifecycleEventArgs $e){
		$entity = $e->getEntity();
		if($this->isEntityTaskCalendarInterface($entity)){
			$this->taskQueueProducer->sendProjectCalendar($entity);
            /** @var \Service\ProjectLog\ProjectLog $serviceProjectLog */
            $serviceProjectLog = InstanceManager::getInstance()->getService('\Service\ProjectLog');
            if($entity->getProject()->getProjectStatus() && $entity->getProject()->getProjectStatus()->getGeneratesLog() == 'Y') {
                $serviceProjectLog->createLogInsert($entity);
                $this->needsFlush = true;
            }

            $this->logProject = array();

		}
        if($this->isEntityAllocation($entity)){
            if($entity->getTask()->getProject()->getProjectStatus() && $entity->getTask()->getProject()->getProjectStatus()->getGeneratesLog() == 'Y') {
                /** @var \Service\ProjectLog\ProjectLog $serviceProjectLog */
                $serviceProjectLog = InstanceManager::getInstance()->getService('\Service\ProjectLog');
                $serviceProjectLog->createLogInsertAllocation($entity);
                $this->needsFlush = true;
            }
        }

	}

    /***********************************************************************
     * Doctrine event. Called after a data is removed from database.
     ***********************************************************************/
	public function postRemove(LifecycleEventArgs $e){
		$entity = $e->getEntity();
        if($entity instanceof Allocation ){
            if($entity->getTask()->getProject()->getProjectStatus() && $entity->getTask()->getProject()->getProjectStatus()->getGeneratesLog() == 'Y') {
                /** @var \Service\ProjectLog\ProjectLog $serviceProjectLog */
                $serviceProjectLog = InstanceManager::getInstance()->getService('\Service\ProjectLog');
                $serviceProjectLog->createLogRemoveAllocation($entity);
                $this->needsFlush = true;
            }
        }
	}

    /***********************************************************************
     * Doctrine event. Called before a data is updated in a database.
     ***********************************************************************/
	public function preUpdate(PreUpdateEventArgs $e){
        $entity = $e->getEntity();
        if($this->isEntityTaskCalendarInterface($entity)) {
            $entity->setInCalc('Y');
            if ($e->hasChangedField("project")) {
                $this->oldProject = $e->getOldValue("project");
                $this->oldProject->setInCalc('Y');
                $this->updateEntity($entity, $e);
            }else if($this->isEntityTask($entity)){
                $entity->getProject()->setInCalc('Y');
                if($entity->getParent()){
                    //$entity->getParent()->setInCalc('Y');
                }
            }
            if ($e->hasChangedField("parent")) {
                $this->oldParent = $e->getOldValue("parent");
            } else {
                $this->oldParent = $e->getEntity()->getParent();
            }
        }
        $this->projectLog($e, $entity);

        if($this->isEntityTask($entity) && $e->hasChangedField("project")) {
            $entity->setCreatedAtInProject(new \DateTime());
        }
    }

    /***********************************************************************
     * Doctrine event. Called after a data is updated in a database.
     ***********************************************************************/
    public function postUpdate(LifecycleEventArgs $e) {
		$entity = $e->getEntity();
		if ($this->isEntityTaskCalendarInterface($entity)) {
			if ($this->oldProject) {
				$this->taskQueueProducer->sendProjectCalendar($this->oldProject);
			}
			if ($this->oldParent != $entity->getParent()) {
				if($this->oldParent) {
					$this->taskQueueProducer->sendProjectCalendar($this->oldParent);
				}
				$this->updateEntity($entity, $e);
			}
			$this->taskQueueProducer->sendProjectCalendar($entity);
			if($entity->getTreeParent()) {
				$this->taskQueueProducer->sendProjectCalendar($entity->getTreeParent());
			}
		} else if ($this->isEntityCalendarDate($entity)){
            PmCache::getInstance()->Clear();
        }
        if(!empty($this->logProject) && $this->isEntityProjectLogInterface($entity)){
            /** @var \Service\ProjectLog\ProjectLog $serviceProjectLog */
            $serviceProjectLog = InstanceManager::getInstance()->getService('\Service\ProjectLog');
            $this->needsFlush = true;
            foreach ($this->logProject as $logProject){
                $serviceProjectLog->createLogUpdate($entity,$logProject['OLD_VALUE'],$logProject['NEW_VALUE'],$logProject['COLUMN']);
            }
            $this->logProject = array();
        }
	}

    /***********************************************************************
     * Doctrine event. Called after a data is flushed in database.
     ***********************************************************************/
    public function postFlush(PostFlushEventArgs $e) {
        if($this->needsFlush) {
            $this->needsFlush = false;
            $e->getEntityManager()->flush();
        }
    }

	/***********************************************************************
	 * Gets registered events.
	 ***********************************************************************/
	public function getSubscribedEvents() {
		return array(
			Events::postUpdate,
			Events::prePersist,
			Events::postPersist,
			Events::preUpdate,
            Events::postRemove,
            Events::postFlush
		);
	}

    /**
     * @param $parent
     * @param $tableName
     * @param PreUpdateEventArgs $e
     * @return string
     */
	private function nextId($parent, $tableName,$e){
		if(!empty($parent)) {
			$tree_id = $parent->getTreeCode();
		} else {
			$tree_id = 0;
		}
		$sql = sprintf(self::SQL_TOCHILDREN, $tableName);
		$ids = $e->getObjectManager()->getConnection()->fetchAll($sql, array("KEY" => $tree_id.".%", "KEY_PARENT" => $tree_id.".%.%" )) ;
		$key_ids = array();
		foreach ($ids as $key => $id) {
			$parts = explode(".",$id["TREE_KEY"]);
			$key_ids[$key] = end($parts);
		}
		$maxId = count($key_ids) > 0 ?  max($key_ids)+1 : 1;
		return $tree_id.".".$maxId;
	}

    /***********************************************************************
     * Check if project status generates log.
     ***********************************************************************/
    private function isProjectStatusLogGenerator($projectStatus){
        return ($projectStatus && $projectStatus->getGeneratesLog() == 'Y');
    }

    /***********************************************************************
     * Check if project generates logs.
     ***********************************************************************/
    private function doesProjectGenerateLog($project){
        return ($project->getProjectStatus() && $this->isProjectStatusLogGenerator($project->getProjectStatus()));
    }

    /***********************************************************************
     * Mounts log value array
     ***********************************************************************/
    private function createLogStructure($oldValue, $newValue, $column){
        return array(
            'OLD_VALUE' => $oldValue,
            'NEW_VALUE' => $newValue,
            'COLUMN'    => $column
        );
    }

    /***********************************************************************
     * Add log to project's property
     ***********************************************************************/
    private function addProjectLog($oldValue, $newValue, $column){
        array_push($this->logProject, $this->createLogStructure($oldValue, $newValue, $column));
    }

    /**
     * @param PreUpdateEventArgs $e
     * @param ProjectLogInterface $entity
     */
    protected function projectLog(PreUpdateEventArgs $e, $entity)
    {
        if($this->isEntityProjectLogInterface($entity)){

            if ($e->hasChangedField("projectStatus")){
                $oldValue = $e->getOldValue("projectStatus");
                $newValue = $entity->getProjectStatus();
                if ($this->isProjectStatusLogGenerator($oldValue) || $this->isProjectStatusLogGenerator($newValue)){
                    $newValue = $newValue ? $newValue->getId() : null;
                    $oldValue = $oldValue ? $oldValue->getId() : null;
                    $this->addProjectLog($oldValue, $newValue, 'PROJECT_STATUS');
                }
            }

            if( $this->doesProjectGenerateLog($entity)){

                if ($e->hasChangedField('estimatedBeginDate')) {
                    $oldValue = $e->getOldValue("estimatedBeginDate")->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
                    $newValue = $entity->getEstimatedBeginDate()->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
                    if ($newValue != $oldValue) {
                        $this->addProjectLog($oldValue, $newValue, 'ESTIMATED_BEGIN_DATE');
                    }
                }
                if ($e->hasChangedField('estimatedEndDate')) {
                    $oldValue = $e->getOldValue("estimatedEndDate")->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
                    $newValue = $entity->getEstimatedEndDate()->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
                    if ($newValue != $oldValue) {
                        $this->addProjectLog($oldValue, $newValue, 'ESTIMATED_END_DATE');
                    }
                }
                if ($e->hasChangedField('beginDate') && $this->isEntityProject($entity)) {
                    $oldValue = $e->getOldValue("beginDate")->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
                    $newValue = $entity->getEstimatedBeginDate()->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
                    if ($newValue != $oldValue) {
                        $this->addProjectLog($oldValue, $newValue, 'BEGIN_DATE');
                    }
                }
                if ($e->hasChangedField('endDate') && $this->isEntityProject($entity)) {
                    $oldValue = $e->getOldValue("endDate")->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
                    $newValue = $entity->getEstimatedEndDate()->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
                    if ($newValue != $oldValue) {
                        $this->addProjectLog($oldValue, $newValue, 'END_DATE');
                    }
                }
                if ($e->hasChangedField('duration')) {
                    /** @var \Service\Calendar\TaskCalendar $utilities */
                    $taskCalendar = InstanceManager::getInstance()->getService('\Service\Calendar\TaskCalendar');
                    $oldValue = $e->getOldValue("duration");
                    $newValue = $entity->getDuration();
                    if (!ctype_digit($oldValue)) {
                        $oldValue = $taskCalendar->getDurationInMilliseconds($oldValue, $entity);
                    }
                    if ($oldValue != $newValue) {
                        $this->addProjectLog($oldValue, $newValue, 'DURATION');
                    }
                }
                if ($e->hasChangedField('timePlanned')) {
                    /** @var \Service\Calendar\TaskCalendar $utilities */
                    $taskCalendar = InstanceManager::getInstance()->getService('\Service\Calendar\TaskCalendar');
                    $oldValue = $e->getOldValue("timePlanned");
                    $newValue = $entity->getPlanned();
                    if (!ctype_digit($oldValue)) {
                        $oldValue = $taskCalendar->getDurationInMilliseconds($oldValue, $entity);
                    }
                    if ($oldValue != $newValue) {
                        $this->addProjectLog($oldValue, $newValue, 'TIME_PLANNED');
                    }
                }
                if ($e->hasChangedField('status')) {
                    $oldValue = $e->getOldValue("status");
                    $newValue = $entity->getStatus();
                    $this->addProjectLog($oldValue, $newValue, 'STATUS');
                }
                if ($e->hasChangedField('project')) {
                    $oldValue = $e->getOldValue("project")->getId();
                    $newValue = $entity->getProject()->getId();
                    $this->addProjectLog($oldValue, $newValue, 'PROJECT');
                }
            }
        }
    }


}
